# Tcl package index file, version 1.0
# This file contains package information for Windows-specific extensions.
#
# Copyright (c) 1997 by Sun Microsystems, Inc.
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# 
# SCCS: @(#) pkgIndex.tcl 1.3 97/12/19 10:49:30

proc loadJava {dir} {
    if {![info exist ::env(CLASSPATH)]} {
	set ::env(CLASSPATH) ""
    } 
    set oldpath $::env(CLASSPATH)
    set ::env(CLASSPATH) "[file join $dir tclblend.jar];$::env(CLASSPATH)"
    load [file join $dir tclblend.dll]
    set ::env(CLASSPATH) $oldpath
    rename loadJava {}
}

package ifneeded java 1.0 [list loadJava $dir]
